# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
from datetime import datetime
from time import sleep
from abc import ABC, abstractmethod

from chipscopy.dm import NodeListener
from chipscopy.dm.harden.noc_perfmon.noc_types import (
    noc_nmu_typedef,
    noc_nsu_typedef,
    ddrmc_noc_typedef,
    noc_node_types,
    ddrmc_main_typedef,
    hbmmc_typedef,
    ddrmc_crypto_typedef,
)
from chipscopy.utils.logger import log

# %%
DOMAIN = "noc_perfmon"
log.register_domain(DOMAIN)
epoch = datetime.utcfromtimestamp(0)  # this is stupid, be best python

# %%
# defines
MAX_SAMPLES = 50
ch_map = {"r": 0, "w": 1}


# %%
def get_noc_typedef_from_name(name) -> str:
    nmu_re = re.compile(".*NMU.*", flags=re.IGNORECASE)
    nsu_re = re.compile(".*NSU.*", flags=re.IGNORECASE)
    ddrmc_noc_re = re.compile(r".*DDRMC\d*_NOC.*", flags=re.IGNORECASE)
    ddrmc_main_re = re.compile(".*DDRMC_MAIN.*", flags=re.IGNORECASE)
    ddrmc_re = re.compile(r".*DDRMC\d*_(S\d+)?X.*", flags=re.IGNORECASE)
    hbmmc_re = re.compile(".*HBM_MC_X.*", flags=re.IGNORECASE)
    ddrmc_crypto_re = re.compile(".*DDRMC_CRYPTO.*", flags=re.IGNORECASE)

    if nmu_re.match(name):
        return noc_nmu_typedef
    elif nsu_re.match(name):
        return noc_nsu_typedef
    elif ddrmc_noc_re.match(name):
        return ddrmc_noc_typedef
    elif ddrmc_main_re.match(name) or ddrmc_re.match(name):
        return ddrmc_main_typedef
    elif hbmmc_re.match(name):
        return hbmmc_typedef
    elif ddrmc_crypto_re.match(name):
        return ddrmc_crypto_typedef
    else:
        log[DOMAIN].debug(f"unknown type for name: {name}")
        return ""


def decode_ddrmc_gen(name):
    ddrmc5_re = re.compile(".*DDRMC5_X.*", flags=re.IGNORECASE)
    if ddrmc5_re.match(name):
        return 5
    else:
        return 4


def get_nsu_count_per_ddrmc(ddrmc_gen):
    if ddrmc_gen == 5:
        return 2
    else:
        return 4


# %%
class NoCElement(ABC):
    """
    Goal of this class is to serve as storage for the data in nodeUpdate events
    """

    def __init__(
        self,
        name,
        alt_name,
        num_samples,
        sampling_period_ms,
        record_to_file,
        tslide=0,
        counter_freq_mhz=1000.0,
    ):
        self.name = name
        self.alt_name = alt_name
        self.num_samples = num_samples
        self.sampling_period_ms = sampling_period_ms
        self.record_to_file = record_to_file
        self.node_type = get_noc_typedef_from_name(self.name)
        self.tslide = tslide
        self.counter_freq_mhz = counter_freq_mhz

        self.samples = {
            "read_bandwidth": [0] * MAX_SAMPLES,
            "write_bandwidth": [0] * MAX_SAMPLES,
            "avg_read_latency": [0] * MAX_SAMPLES,
            "avg_write_latency": [0] * MAX_SAMPLES,
            "flags": [],
            "ts": [],
            "type": self.node_type,
        }
        self.raw_trace_data = {}

    @abstractmethod
    def update_node(self, tcf_node, updated_keys):  # pragma: no cover
        pass

    def get_axis_metrics(self, axis, view, pc):
        # pc is only for HBM
        for metric in self.axis_metrics[axis][view]:
            yield metric

    def record_bw_lat(
        self,
        read_bytes,
        write_bytes,
        read_bytes_per_s,
        write_bytes_per_s,
        avg_read_latency,
        avg_write_latency,
    ):
        self.samples["read_bandwidth"].append(read_bytes_per_s)
        self.samples["write_bandwidth"].append(write_bytes_per_s)
        self.samples["avg_read_latency"].append(avg_read_latency)
        self.samples["avg_write_latency"].append(avg_write_latency)
        log[DOMAIN].info(
            f"{self.name}: rb: {read_bytes}, wb: {write_bytes}, "
            + f"rbw: {read_bytes_per_s}, wbw: {write_bytes_per_s}, "
            + f"arl: {avg_read_latency}, awl: {avg_write_latency}"
        )

    def trim_and_log(self, raw_trace_data):
        # trim samples down:
        for _, sample_data in self.samples.items():
            if len(sample_data) > MAX_SAMPLES:
                sample_data.pop(0)

        if self.record_to_file:
            date_handler = lambda obj: (
                obj.isoformat()
                if isinstance(obj, (datetime, datetime.date))
                else json.JSONEncoder().default(obj)
            )
            with open("raw_trace_data.json", "a") as ofh:
                json.dump(raw_trace_data, ofh, default=date_handler)
                ofh.write("\n")
            with open("scaled_trace_data.json", "a") as ofh:
                data_to_dump = {self.name: self.samples}
                json.dump(data_to_dump, ofh, default=date_handler)
                ofh.write("\n")

        self.raw_trace_data = raw_trace_data


class DDRMC(NoCElement):
    dc_metrics_map = {
        0: "activates",
        1: "read_cas",
        2: "write_cas",
        3: "precharge",
        4: "precharge_all",
        5: "refresh",
        6: "idle",
        7: "overhead",
        8: "bus_turnarounds",
    }

    def __init__(self, *args):
        """
        for DDRMC 'ddrmc_main_X' shall be the name and 'ddrmc_noc_X' shall serve as the alternate name
        :param name:
        :param alt_name:
        """
        super().__init__(*args)
        self.num_nsu = get_nsu_count_per_ddrmc(decode_ddrmc_gen(self.name))
        # add DC aggregate, NSU agg, and NSU port metrics
        for ch in range(0, 2):
            for metric in DDRMC.dc_metrics_map.values():
                self.samples.update({f"dc{ch}_{metric}": [0] * MAX_SAMPLES})
                if ch == 0:
                    self.samples.update({f"agg_{metric}": [0] * MAX_SAMPLES})
            self.samples.update({f"dc{ch}_flags": [0] * MAX_SAMPLES})

        for nsu in range(0, self.num_nsu):
            self.samples.update({f"nsu{nsu}_write_bandwidth": [0] * MAX_SAMPLES})
            self.samples.update({f"nsu{nsu}_read_bandwidth": [0] * MAX_SAMPLES})
            self.samples.update({f"nsu{nsu}_write_latency": [0] * MAX_SAMPLES})
            self.samples.update({f"nsu{nsu}_read_latency": [0] * MAX_SAMPLES})

        self.axis_metrics = {
            "left_axis": {
                "bandwidth": ["write_bandwidth", "read_bandwidth"],
                "latency": ["avg_write_latency", "avg_read_latency"],
                "mem": [
                    "agg_activates",
                    "agg_read_cas",
                    "agg_write_cas",
                    "agg_precharge",
                    "agg_precharge_all",
                    "agg_refresh",
                    "agg_idle",
                    "agg_overhead",
                    "agg_bus_turnarounds",
                ],
            },
            "right_axis": {"bandwidth": [], "latency": [], "mem": []},
        }

    def update_node(self, tcf_node, updated_keys):
        self.num_samples -= 1
        raw_trace_data = {
            "type": self.node_type,
            "name": self.name,
            # 'ts': (datetime.now() - self.start_time) + epoch,
            "ts": datetime.now(),
            "r": {},
            "w": {},
        }

        tcf_node_type = get_noc_typedef_from_name(tcf_node.Name)

        if tcf_node_type == ddrmc_noc_typedef:
            for ch, mon in ch_map.items():
                raw_trace_data[ch]["lacc"] = []
                raw_trace_data[ch]["burst_count"] = []
                raw_trace_data[ch]["byte_count"] = []
                raw_trace_data[ch]["flags"] = []

                for nsu in range(0, self.num_nsu):
                    if decode_ddrmc_gen(self.name) == 5:
                        lacc = tcf_node[f"nsu{nsu}_perf_mon_{mon}_0"]
                        lacc += (tcf_node[f"nsu{nsu}_perf_mon_{mon}_1"] & 0xFFFF) << 32
                        raw_trace_data[ch]["lacc"].append(lacc)
                        bc = tcf_node[f"nsu{nsu}_perf_mon_{mon}_2"]
                        bc += (tcf_node[f"nsu{nsu}_perf_mon_{mon}_3"] & 0xFFFF) << 32
                        raw_trace_data[ch]["burst_count"].append(bc)
                        hfc = tcf_node[f"nsu{nsu}_perf_mon_{mon}_4"]
                        hfc += (tcf_node[f"nsu{nsu}_perf_mon_{mon}_5"] & 0xFFFF) << 32
                        raw_trace_data[ch]["byte_count"].append(hfc)
                        raw_trace_data[ch]["flags"].append((lacc & 0x80000000) >> 31)
                    else:
                        lacc = tcf_node[f"nsu{nsu}_perf_mon_{mon}_0"]
                        raw_trace_data[ch]["lacc"].append(lacc & 0x7FFFFFFF)
                        raw_trace_data[ch]["burst_count"].append(
                            tcf_node[f"nsu{nsu}_perf_mon_{mon}_1"]
                        )
                        raw_trace_data[ch]["byte_count"].append(
                            tcf_node[f"nsu{nsu}_perf_mon_{mon}_2"]
                        )
                        raw_trace_data[ch]["flags"].append((lacc & 0x80000000) >> 31)

                    if lacc == None:  # why does this logic control exist?
                        continue

            # bandwidth
            rf = 0
            wf = 0
            for nsu in range(0, self.num_nsu):
                rf += raw_trace_data["r"]["burst_count"][nsu]
                wf += raw_trace_data["w"]["burst_count"][nsu]
            read_bytes = rf * 128 / 8
            write_bytes = wf * 128 / 8
            read_bytes_per_s = read_bytes / (self.sampling_period_ms / 1000)
            write_bytes_per_s = write_bytes / (self.sampling_period_ms / 1000)

            # flags
            flags = []
            for ch in ["r", "w"]:
                flag = 0
                for nsu in range(0, self.num_nsu):
                    flag += raw_trace_data[ch]["flags"][nsu]
                if flag != 0:
                    flags.append(f"{ch}_overflow")
            self.samples["flags"].append(flags)

            # latency
            avg_lat = {"r": 0, "w": 0}
            for ch in ["r", "w"]:
                agg_lat = 0
                agg_burst = 0
                for nsu in range(0, self.num_nsu):
                    agg_lat += raw_trace_data[ch]["lacc"][nsu]
                    agg_burst += raw_trace_data[ch]["burst_count"][nsu]
                if agg_burst == 0:
                    agg_burst = 1
                avg_lat[ch] = float(agg_lat) / float(agg_burst)

            # common op
            self.record_bw_lat(
                read_bytes,
                write_bytes,
                read_bytes_per_s,
                write_bytes_per_s,
                avg_lat["r"],
                avg_lat["w"],
            )

        if tcf_node_type == ddrmc_main_typedef:
            total_ops = 0
            disagg_data = {}
            for ch in range(0, 2):
                flags = 0
                for m_index, metric in DDRMC.dc_metrics_map.items():
                    agg_metric = f"agg_{metric}"
                    disagg_metric = f"dc{ch}_{metric}"
                    self.samples[disagg_metric].append(0)
                    if ch == 0:
                        self.samples[agg_metric].append(0)
                    data = tcf_node[f"dc{ch}_perf_mon_{m_index}"]
                    raw_trace_data[disagg_metric] = data
                    disagg_data[disagg_metric] = data
                    overflow = data & 0x80000000 == 0x80000000
                    data &= 0x7FFF_FFFF
                    if overflow:
                        flags |= 1 << m_index  # overflow
                    self.samples[disagg_metric][-1] = data
                    self.samples[agg_metric][-1] += data
                    total_ops += data
                    # need to trim here or we can get a dim mismatch when plotting happens
                    # unique issue for ddmrc, since it's two nodes with polls coming in at different rates
                    if len(self.samples[disagg_metric]) > MAX_SAMPLES:
                        self.samples[disagg_metric].pop(0)
                    if ch == 1:
                        if len(self.samples[agg_metric]) > MAX_SAMPLES:
                            self.samples[agg_metric].pop(0)
                self.samples[f"dc{ch}_flags"].append(flags)
            for m_index, metric in DDRMC.dc_metrics_map.items():
                agg_metric = f"agg_{metric}"
                # disagg_metric = f"dc{ch}_{metric}"
                # converting this to float now, it's a 100
                if total_ops > 0:
                    self.samples[agg_metric][-1] = (self.samples[agg_metric][-1] * 100) / total_ops

            log[DOMAIN].info(disagg_data)

        # timestamp handling
        # the way the polls are created the NA reports AFTER the main, so when this event is received, update plots
        if tcf_node_type == ddrmc_noc_typedef:
            time_delta = raw_trace_data["ts"].strftime("%M:%S.") + str(
                int(raw_trace_data["ts"].microsecond / 1000)
            )
            self.samples["ts"].append(time_delta)
            self.trim_and_log(raw_trace_data)


class NoCMasterSlave(NoCElement):
    def __init__(self, *args):
        super().__init__(*args)
        self.samples.update(
            {
                "min_r_latency": [0] * MAX_SAMPLES,
                "max_r_latency": [0] * MAX_SAMPLES,
                "min_w_latency": [0] * MAX_SAMPLES,
                "max_w_latency": [0] * MAX_SAMPLES,
            }
        )
        self.axis_metrics = {
            "left_axis": {
                "bandwidth": ["write_bandwidth", "read_bandwidth"],
                "latency": [
                    "avg_write_latency",
                    "avg_read_latency",
                    "min_w_latency",
                    "max_w_latency",
                    "min_r_latency",
                    "max_r_latency",
                ],
                "mem": [],
            },
            "right_axis": {"bandwidth": [], "latency": [], "mem": []},
        }

    def update_node(self, tcf_node, updated_keys):
        raw_trace_data = {
            "type": self.node_type,
            "name": self.name,
            # 'ts': (datetime.now() - self.start_time) + epoch,
            "ts": datetime.now(),
            "r": {},
            "w": {},
        }
        for ch, mon in ch_map.items():
            raw_trace_data[ch]["burst_count"] = tcf_node[f"reg_perf_mon{mon}_burst_cnt"]
            raw_trace_data[ch]["byte_count"] = tcf_node[f"reg_perf_mon{mon}_byte_cnt_lwr"] + (
                (tcf_node[f"reg_perf_mon{mon}_cnt_and_ofl"] & 0xFF) << 32
            )
            raw_trace_data[ch]["flags"] = (tcf_node[f"reg_perf_mon{mon}_cnt_and_ofl"] & 0x1F00) >> 8
            raw_trace_data[ch]["lacc"] = (
                tcf_node[f"reg_perf_mon{mon}_latency_acc_lwr"]
                + ((tcf_node[f"reg_perf_mon{mon}_latency_acc_upr"] & 0xFF) << 32)
            ) << self.tslide
            raw_trace_data[ch]["lmax"] = tcf_node[f"reg_perf_mon{mon}_latency_max"] << self.tslide
            raw_trace_data[ch]["lmin"] = tcf_node[f"reg_perf_mon{mon}_latency_min"] << self.tslide

        # bandwidth
        read_bytes = raw_trace_data["r"]["byte_count"]
        write_bytes = raw_trace_data["w"]["byte_count"]
        read_bytes_per_s = read_bytes * 1000 / self.sampling_period_ms
        write_bytes_per_s = write_bytes * 1000 / self.sampling_period_ms
        # latency and flags
        flags = []
        for ch in ["r", "w"]:
            self.samples[f"min_{ch}_latency"].append(raw_trace_data[ch]["lmin"])
            self.samples[f"max_{ch}_latency"].append(raw_trace_data[ch]["lmax"])
            # flags
            raw_flags = raw_trace_data[ch]["flags"]
            if raw_flags != 0:
                if (raw_flags & 0x10) == 0x10:
                    flags.append(f"{ch}_byte_count_overflow")
                if (raw_flags & 0x08) == 0x08:
                    flags.append(f"{ch}_burst_count_overflow")
                if (raw_flags & 0x04) == 0x04:
                    flags.append(f"{ch}_accumulated_latency_overflow")
                if (raw_flags & 0x02) == 0x02:
                    flags.append(f"{ch}_max_latency_overflow")
                if (raw_flags & 0x01) == 0x01:
                    flags.append(f"{ch}_min_latency_overflow")
        self.samples["flags"].append(flags)

        # latency
        avg_lat = {"r": 0, "w": 0}
        for ch in ["r", "w"]:
            avg_lat[ch] = raw_trace_data[ch]["lacc"] / (
                raw_trace_data[ch]["burst_count"] if raw_trace_data[ch]["burst_count"] != 0 else 1
            )
        self.record_bw_lat(
            read_bytes, write_bytes, read_bytes_per_s, write_bytes_per_s, avg_lat["r"], avg_lat["w"]
        )
        time_delta = raw_trace_data["ts"].strftime("%M:%S.") + str(
            int(raw_trace_data["ts"].microsecond / 1000)
        )
        self.samples["ts"].append(time_delta)
        self.trim_and_log(raw_trace_data)


class NoCPerfMonNodeListener(NodeListener):
    """
    Example Listener class that will print out metrics from NoC nodes (NMU|NSU|DDRMC)

    """

    def __init__(
        self,
        sampling_period_ms,
        num_samples,
        enabled_nodes,
        record_to_file,
        npi_clk_freq=299997009,
        noc_clk_freq=999989990,
        extended_monitor_config=None,
    ):
        self.sampling_period_ms = sampling_period_ms
        self.num_samples = {}
        self.npi_clk_freq = npi_clk_freq
        self.noc_clk_freq = noc_clk_freq
        self.plotter = None
        super().__init__()
        # setup logging
        log.enable_domain([DOMAIN])
        self.level = "WARNING"
        log.change_log_level("WARNING")
        # self.start_time = datetime.now()
        self.noc_elements = {}  # storage for the active monitors
        self.unique_elements = (
            {}
        )  # same objects from above, but with alt mapping removed -- for plot utils

        # build data structure for holding historical data
        for elem in enabled_nodes:
            tslide = 0
            if extended_monitor_config is not None:
                if elem in extended_monitor_config.keys():
                    node_config = extended_monitor_config[elem]
                    if "tslide" in node_config:
                        tslide = node_config["tslide"]
            node_type = get_noc_typedef_from_name(elem)
            if node_type in [noc_nmu_typedef, noc_nsu_typedef]:
                storage_node = NoCMasterSlave(
                    elem.lower(),
                    None,
                    num_samples,
                    sampling_period_ms["NPI"],
                    record_to_file,
                    tslide,
                    npi_clk_freq / 1000000.0,
                )
                self.noc_elements[elem.lower()] = storage_node
                self.unique_elements[elem.lower()] = storage_node
            elif node_type in [ddrmc_main_typedef]:
                alt_name = "ddrmc_noc_" + elem.split("_")[-1]
                node = DDRMC(
                    elem.lower(),
                    alt_name.lower(),
                    num_samples,
                    sampling_period_ms["NoC"],
                    record_to_file,
                    tslide,
                    noc_clk_freq / 1000000.0,
                )
                self.noc_elements[elem.lower()] = node
                self.noc_elements[alt_name.lower()] = node
                self.unique_elements[elem.lower()] = node
                # TODO iff gen5
                # TODO node names are lacking the '5' fixup
                crypto_name = "ddrmc_crypto_" + elem.split("_")[-1]
                crypto_node = DDRMCCrypto(
                    crypto_name.lower(), sampling_period_ms[elem], record_to_file
                )
                self.noc_elements[crypto_name.lower()] = crypto_node
                # self.unique_elements[elem.lower()] = node
            elif node_type in [hbmmc_typedef]:
                from chipscopy.api.noc.graphing.hbmmc import HBMMC

                node = HBMMC(
                    elem.lower(),
                    elem.lower(),
                    num_samples,
                    sampling_period_ms[elem],
                    record_to_file,
                    tslide,
                    npi_clk_freq / 1000000.0,
                )
                self.noc_elements[elem.lower()] = node
                self.unique_elements[elem.lower()] = node
            else:  # pragma: no cover
                raise Exception(f"Error, unsupported node: {elem}, type: {node_type}")

    def node_changed(self, node, updated_keys):
        node_type = get_noc_typedef_from_name(node.Name)
        name = node.Name.lower()
        if node_type in noc_node_types and node.Name.lower() in self.noc_elements.keys():
            log[DOMAIN].debug(f"{node.type}: {node.Name}")

            # dispatch the class handler
            self.noc_elements[name].update_node(node, updated_keys)

            # finally update plots
            if self.plotter is not None and node_type is not ddrmc_main_typedef:
                self.plotter.update(self.noc_elements[name])

    def link_plotter(self, plotter):
        self.plotter = plotter
        self.plotter.listener = self

    def change_log_level(self, new_level):
        self.level = new_level
        log.change_log_level(self.level)


class PerfTGController:
    """
    This class is designed to aid in controlling and using the Performance AXI Traffic Generator LogicCore IP for
    Versal series ACAP devices from AMD. It supports no other architectures or Traffic Generator IPs. There are several
    others in the default catalog.


    """

    def __init__(self, base_address, device, vio=None):
        self.base_address = base_address
        self.device = device
        self.vio = vio
        self.instr_ba = self.base_address + 0x8000
        self.ctrl = self.base_address + 0x4000
        self.tg_start = self.base_address + 0x4004
        self.instr_0 = []
        self.instr_1 = []
        self.supported_patterns = {
            "Write Linear": [
                [
                    0xFF200000,
                    0x00080000,
                    0x00000002,
                    0x00000000,
                    0xFFE00000,
                    0x000FFFFF,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00010020,
                    0x00143500,
                    0x0000048C,
                    0x0,
                ],
                [
                    0x00000001,
                    0x00100000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00020000,
                    0x00000000,
                    0x00000000,
                    0x0,
                ],
            ],
            "Read Linear": [
                [
                    0xFF200000,
                    0x00000000,
                    0x00000002,
                    0x00000000,
                    0xFFE00000,
                    0x000FFFFF,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00010020,
                    0x00143500,
                    0x0000048E,
                    0x0,
                ],
                [
                    0x00000001,
                    0x00100000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00020000,
                    0x00000000,
                    0x00000000,
                    0x0,
                ],
            ],
        }
        self.active_pattern = "Write Linear"
        self.trigger_block_rst_probe = None
        self.tg_rst_probe = None
        for probe in self.vio.probe_names:
            if probe.split("/")[-1] == "noc_sim_trig_rst_n":
                self.trigger_block_rst_probe = probe
            if probe.split("/")[-1] == "noc_tg_tg_rst_n":
                self.tg_rst_probe = probe

    def start(self):
        self.device.memory_write(self.tg_start, [0x1])
        # CR-1062874
        # Second write of the ctrl reg is required for the start command to be effective.
        sleep(0.5)
        self.device.memory_write(self.tg_start, [0x1])

    def stop(self):
        # this doesn't seem to be doing much
        self.device.memory_write(self.tg_start, [0x0])

        self._soft_reset()
        self._clear_instruction_memory()

    def block_reset(self):
        if self.vio is not None:
            self.vio.write_probes(
                {
                    self.trigger_block_rst_probe: 0x0,
                    self.tg_rst_probe: 0x0,
                }
            )
            self.vio.write_probes(
                {
                    self.trigger_block_rst_probe: 0x1,
                    self.tg_rst_probe: 0x1,
                }
            )

    def program(self):
        self._soft_reset()
        self.device.memory_write(self.instr_ba, self.supported_patterns[self.active_pattern][0])
        self.device.memory_write(
            self.instr_ba + 0x40, self.supported_patterns[self.active_pattern][1]
        )

    def set_active_pattern(self, pattern):
        if pattern not in self.supported_patterns.keys():
            raise ValueError(f"Unsupported pattern: {pattern}")
        else:
            self.active_pattern = pattern

    def _clear_instruction_memory(self):
        self.device.memory_write(self.instr_ba, [0x0] * 13)
        self.device.memory_write(self.instr_ba + 0x40, [0x0] * 13)

    def _soft_reset(self):
        self.device.memory_write(self.ctrl, [0x0])
        self.device.memory_write(self.ctrl, [0x1])


class DDRMCCrypto:
    def __init__(self, name, num_samples, record_to_file):
        self.name = name
        self.node_type = get_noc_typedef_from_name(self.name)
        self.num_samples = num_samples
        self.record_to_file = record_to_file

    def update_node(self, tcf_node, updated_keys):
        self.aes_xts_enc = tcf_node["reg_crypto_perf_mon_cnt0"]
        self.aes_xts_dec = tcf_node["reg_crypto_perf_mon_cnt1"]
        self.gcm_enc = tcf_node["reg_crypto_perf_mon_cnt2"]
        self.gcm_dec = tcf_node["reg_crypto_perf_mon_cnt3"]
        self.pwr_cache_rd_hit = tcf_node["reg_crypto_perf_mon_pwr0"]
        self.pwr_cache_rd_miss = tcf_node["reg_crypto_perf_mon_pwr1"]
        self.pwr_cache_wr_hit = tcf_node["reg_crypto_perf_mon_pwr2"]
        self.pwr_cache_wr_miss = tcf_node["reg_crypto_perf_mon_pwr3"]
        self.pwr_cache_evict = tcf_node["reg_crypto_perf_mon_pwr4"]
        self.mc_rd_hit = tcf_node["reg_crypto_perf_mon_mc0"]
        self.mc_rd_miss = tcf_node["reg_crypto_perf_mon_mc1"]
        self.mc_wr_hit = tcf_node["reg_crypto_perf_mon_mc2"]
        self.mc_wr_miss = tcf_node["reg_crypto_perf_mon_mc3"]
        self.mc_evict = tcf_node["reg_crypto_perf_mon_mc4"]

        raw_trace_data = {
            "type": self.node_type,
            "name": self.name,
            "ts": datetime.now(),
            "aes_xts_enc": tcf_node["reg_crypto_perf_mon_cnt0"],
            "aes_xts_dec": tcf_node["reg_crypto_perf_mon_cnt1"],
            "gcm_enc": tcf_node["reg_crypto_perf_mon_cnt2"],
            "gcm_dec": tcf_node["reg_crypto_perf_mon_cnt3"],
            "pwr_cache_rd_hit": tcf_node["reg_crypto_perf_mon_pwr0"],
            "pwr_cache_rd_miss": tcf_node["reg_crypto_perf_mon_pwr1"],
            "pwr_cache_wr_hit": tcf_node["reg_crypto_perf_mon_pwr2"],
            "pwr_cache_wr_miss": tcf_node["reg_crypto_perf_mon_pwr3"],
            "pwr_cache_evict": tcf_node["reg_crypto_perf_mon_pwr4"],
            "mc_rd_hit": tcf_node["reg_crypto_perf_mon_mc0"],
            "mc_rd_miss": tcf_node["reg_crypto_perf_mon_mc1"],
            "mc_wr_hit": tcf_node["reg_crypto_perf_mon_mc2"],
            "mc_wr_miss": tcf_node["reg_crypto_perf_mon_mc3"],
            "mc_evict": tcf_node["reg_crypto_perf_mon_mc4"],
        }

        # log_str = f"{self.name} - aes_xts_enc: {raw_trace_data['aes_xts_enc']}"
        log[DOMAIN].info(str(self))
        # self.samples["ts"].append(time_delta)
        # self.trim_and_log(raw_trace_data)

    def __str__(self):
        ret_str = (
            f"{self.name}: aes_xts_enc: {self.aes_xts_enc}, aes_xts_dec: {self.aes_xts_dec} "
            f"gcm_enc {self.gcm_enc}, gcm_dec: {self.gcm_dec}"
        )
        return ret_str



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""device_scanner.py -

Scans hw_server and cs_server and organize nodes into the devices.

There is a relationship between nodes across views. The device scanner identifies
the available devices, and help to connect associated nodes in the views together
for any given device.

::

    ----------------------------------------------------------------------------
    -- Versal Node Tree Example
    ----------------------------------------------------------------------------
    jtag (ViewInfo)
    └── whole scan chain (JtagCable)                  jsn-VCK190 FT4232H-80500501b019A
        ├── arm_dap (JtagDevice)                      jsn-VCK190 FT4232H-80500501b019A-6ba00477-0
        └── xcvc1902 (JtagDevice)                     jsn-VCK190 FT4232H-80500501b019A-14ca8093-0

    memory (ViewInfo)
    ├── Versal xcvc1902 (MemoryNode)                  JTAG-jsn-VCK190 FT4232H-80500501b019A-6ba00477-0
    │   ├── RPU (MemoryNode)                          JTAG-jsn-VCK190 FT4232H-80500501b019A-6ba00477-0R
        ...
    │   ├── APU (MemoryNode)                          JTAG-jsn-VCK190 FT4232H-80500501b019A-6ba00477-0A
        ...
    │   ├── PPU (MemoryNode)                          JTAG-jsn-VCK190 FT4232H-80500501b019A-6ba00477-0P1
        ...
    │   ├── PSM (MemoryNode)                          JTAG-jsn-VCK190 FT4232H-80500501b019A-6ba00477-0M
        ...
    │   ├── PMC (Node)                                JTAG-jsn-VCK190 FT4232H-80500501b019A-14ca8093-0
    │   └── PL (Node)                                 JTAG-jsn-VCK190 FT4232H-80500501b019A-14ca8093-0PL
    └── DPC (MemoryNode)                              DPC-dpc-jsn-VCK190 FT4232H-80500501b019A-14ca8093-0-70b435d3-0

    debugcore (ViewInfo)
    ├── Versal xcvc1902 (Node)                        JTAG-jsn-VCK190 FT4232H-80500501b019A-6ba00477-0
    └── DPC xcvc1902 (DebugCoreClientNode)            dpc-jsn-VCK190 FT4232H-80500501b019A-14ca8093-0-70b435d3-0

    chipscope (ViewInfo)
    ├── Versal xcvc1902 (Node)                        [TCP:localhost:3121.debugcore]-JTAG-jsn-VCK190 FT4232H-80500501b019A-6ba00477-0
    └── DPC xcvc1902 (CoreParent)                     [TCP:localhost:3121.debugcore]-dpc-jsn-VCK190 FT4232H-80500501b019A-14ca8093-0-70b435d3-0

    ----------------------------------------------------------------------------
    -- US+ Node Tree Example
    ----------------------------------------------------------------------------

    jtag (ViewInfo)
    └── whole scan chain (JtagCable)                  jsn-JTAG-SMT2NC-210308958001
        └── xcku040 (JtagDevice)                      jsn-JTAG-SMT2NC-210308958001-13822093-0
            └── bscan-switch (JtagDevice)             jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1
                ├── axi (JtagDevice)                  jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS1
                └── axi (JtagDevice)                  jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS2

    memory (ViewInfo)
    └── xcku040 (Node)                                JTAG-jsn-JTAG-SMT2NC-210308958001-13822093-0
        ├── axi (Node)                                JTAG-jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS1
        └── axi (Node)                                JTAG-jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS2

    debugcore (ViewInfo)
    └── Ultrascale xcku040 (DebugCoreClientNode)      jsn-JTAG-SMT2NC-210308958001-13822093-0
        ├── DebugHub (Node)                           jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS1
        │   ├── VIO_0 (Node)                          jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS1-DBGCORE02-00000000
        ...
        └── DebugHub (Node)                           jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS2
            ├── Core_0 (Node)                         jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS2-DBGCOREDEE3-00000000
            ...

    chipscope (ViewInfo)
    └── Ultrascale xcku040 (CoreParent)               [TCP:localhost:3120.debugcore]-jsn-JTAG-SMT2NC-210308958001-13822093-0
        ├── DebugHub (Node)                           [TCP:localhost:3120.debugcore]-jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS1
        │   ├── VIO_0 (Node)                          [TCP:localhost:3120.debugcore]-jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS1-DBGCORE02-00000000
        ...
        └── DebugHub (Node)                           [TCP:localhost:3120.debugcore]-jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS2
            ├── Core_0 (Node)                         [TCP:localhost:3120.debugcore]-jsn-JTAG-SMT2NC-210308958001-13822093-0-BS-1-BS2-DBGCOREDEE3-00000000
            ...
"""
import dataclasses
from collections import defaultdict
from typing import Optional, List, Iterator, Dict, TypeVar

from chipscopy.api.device.device_util import get_node_dna
from chipscopy.client import ServerInfo
from chipscopy.utils.logger import log


@dataclasses.dataclass
class JtagRecord:
    server: str
    view: str
    ctx: str
    name: str
    hier_name: str
    part: str
    arch_name: str
    is_arm_dap: bool
    jtag_index: int
    jtag_cable_ctx: str
    dap_ctx: str = ""
    dna: Optional[int] = None


@dataclasses.dataclass
class MemoryRecord:
    server: str
    view: str
    ctx: str
    name: str
    hier_name: str
    jtag_node_id: str = None
    jtag_node_group: str = None
    is_dap: bool = False
    is_dpc: bool = False
    dna: Optional[int] = None


@dataclasses.dataclass
class DebugcoreRecord:
    server: str
    view: str
    ctx: str
    name: str
    hier_name: str
    is_dap: bool = False
    is_dpc: bool = False
    dna: Optional[int] = None


@dataclasses.dataclass
class ChipscopeRecord:
    server: str
    view: str
    ctx: str
    name: str
    hier_name: str
    is_dap: bool = False
    is_dpc: bool = False
    dna: Optional[int] = None


ViewRecordType = TypeVar(
    "ViewRecordType", JtagRecord, MemoryRecord, DebugcoreRecord, ChipscopeRecord
)


def scan_jtag_view(
    hw_server: ServerInfo, include_dna=True, include_arm_dap=False
) -> Iterator[JtagRecord]:
    """
    Scans the jtag chains of a hw_server, returning a dict record for
    devices in order for each cable. Arm-dap devices are collapsed into the
    device they are associated with (by default).

    Args:
        hw_server: Hardware server
        include_dna: True to include dna (default True)
        include_arm_dap: True to return arm-dap devices (default False)

    Returns:
        Iterator of devices in jtag chain

    """
    log.client.debug(
        f"scan_jtag_view: hw_server={hw_server}, include_dna={include_dna}, include_arm_dap={include_arm_dap}"
    )
    previous_jtag_device_record: Optional[JtagRecord] = None
    view = hw_server.get_view("jtag")
    for jtag_cable in view.get_children():
        log.client.debug(f"jtag_scan: scanning cable {jtag_cable.ctx}")
        jtag_index = 0
        error_msg = ""
        if not jtag_cable.props.get("isActive"):
            continue
        if jtag_cable.props.get("isError"):
            desc = jtag_cable.props.get("Description")
            if desc:
                error_msg = f"error: {desc}"
            else:
                error_msg = "error"
        else:
            status = jtag_cable.props.get("Status")
            if status:
                error_msg = f"ERROR: {jtag_cable.ctx}: {status}"
        if error_msg:
            # There is some kind of cable error - do not read the cables
            log.client.error(error_msg)
            continue

        for jtag_device in view.get_children(jtag_cable):
            jtag_record = JtagRecord(
                server="hw_server",
                view="jtag",
                ctx=jtag_device.ctx,
                name=jtag_device.props.get("Name"),
                hier_name=f"jtag/{jtag_device.props.get('Name')}",
                part=jtag_device.props.get("Name"),
                arch_name=jtag_device.props.get("arch_name"),
                is_arm_dap=jtag_device.props.get("Name") == "arm_dap",
                jtag_index=jtag_index,
                jtag_cable_ctx=jtag_cable.ctx,
            )

            if jtag_record.is_arm_dap is False:
                if previous_jtag_device_record and previous_jtag_device_record.is_arm_dap:
                    # Versal and Zynq
                    jtag_record.dap_ctx = previous_jtag_device_record.ctx
                    if include_dna:
                        jtag_record.dna = get_node_dna(jtag_device)
                elif jtag_record.arch_name in ["kintexuplus", "kintexu", "virtexuplus", "virtexu"]:
                    # Virtex, Kintex US+, no arm-dap
                    jtag_record.dap_ctx = ""
                    if include_dna:
                        jtag_record.dna = get_node_dna(jtag_device)
                log.client.debug(f"jtag_record: idx={jtag_index}, device={jtag_record.ctx}")
                log.client.trace(repr(jtag_record))
                yield jtag_record
            else:
                if include_arm_dap:
                    log.client.debug(f"jtag_record: idx={jtag_index}, ctx={jtag_record.ctx}")
                    log.client.trace(repr(jtag_record))
                    yield jtag_record
                else:
                    log.client.debug(
                        f"jtag_record: idx={jtag_index}, ctx={jtag_record.ctx} (EXCLUDED)"
                    )
                    log.client.trace(repr(jtag_record))

            jtag_index += 1
            previous_jtag_device_record = jtag_record


def scan_memory_view(hw_server: ServerInfo, include_dna=True) -> Iterator[MemoryRecord]:
    """
    Scan and return the top level memory nodes

    Args:
        hw_server:
        include_dna:

    Returns:

    """
    log.client.debug(f"scan_memory_view: hw_server={hw_server}, include_dna={include_dna}")
    view = hw_server.get_view("memory")
    for memory_node in view.get_children():
        name = memory_node.props.get("Name", "")
        is_dpc = name.startswith("DPC")
        memory_record = MemoryRecord(
            server="hw_server",
            view="memory",
            ctx=memory_node.ctx,
            name=name,
            hier_name=f"memory/{memory_node.props.get('Name')}",
            jtag_node_id=memory_node.props.get("JtagNodeID"),
            jtag_node_group=memory_node.props.get("JtagGroup"),
            is_dap=not is_dpc,
            is_dpc=is_dpc,
        )
        if include_dna:
            memory_record.dna = get_node_dna(memory_node)
        log.client.debug(f"memory_record: name={memory_record.name}, ctx={memory_record.ctx}")
        log.client.trace(repr(memory_record))
        yield memory_record


def scan_debugcore_view(hw_server: ServerInfo, include_dna=True) -> Iterator[DebugcoreRecord]:
    """
    Args:
        hw_server:
        include_dna:

    Returns:
    """
    log.client.debug(f"scan_debugcore_view: hw_server={hw_server}, include_dna={include_dna}")
    view = hw_server.get_view("debugcore")
    for debugcore_node in view.get_children():
        name = debugcore_node.props.get("Name", "")
        is_dpc = name.startswith("DPC")
        debugcore_record = DebugcoreRecord(
            server="hw_server",
            view="debugcore",
            ctx=debugcore_node.ctx,
            name=name,
            hier_name=f"debugcore/{debugcore_node.props.get('Name')}",
            is_dap=not is_dpc,
            is_dpc=is_dpc,
        )
        if include_dna:
            debugcore_record.dna = get_node_dna(debugcore_node)
        log.client.debug(
            f"debugcore_record: name={debugcore_record.name}, ctx={debugcore_record.ctx}"
        )
        log.client.trace(repr(debugcore_record))
        yield debugcore_record


def scan_chipscope_view(cs_server: ServerInfo, include_dna=True) -> Iterator[ChipscopeRecord]:
    """
    Args:
        cs_server:
        include_dna:

    Returns:
    """
    log.client.debug(f"scan_chipscope_view: cs_server={cs_server}, include_dna={include_dna}")
    view = cs_server.get_view("chipscope")
    for chipscope_node in view.get_children():
        name = chipscope_node.props.get("Name", "")
        is_dpc = name.startswith("DPC")
        chipscope_record = ChipscopeRecord(
            server="cs_server",
            view="chipscope",
            ctx=chipscope_node.ctx,
            name=name,
            hier_name=f"chipscope/{chipscope_node.props.get('Name')}",
            is_dap=not is_dpc,
            is_dpc=is_dpc,
        )
        if include_dna:
            chipscope_record.dna = get_node_dna(chipscope_node)
        log.client.debug(
            f"chipscope_record: name={chipscope_record.name}, ctx={chipscope_record.ctx}"
        )
        log.client.trace(repr(chipscope_record))
        yield chipscope_record


def scan_all_views(
    hw_server: ServerInfo, cs_server: ServerInfo, include_dna=True
) -> Dict[str, List[ViewRecordType]]:
    def populate_dna_or_context_dicts(rec_itr, all_view_dict_, view_dict_):
        # Helper to reduce duplication - just add nodes to the correct dict if dna may exist
        for rec in rec_itr:
            if rec.dna:
                all_view_dict_[str(rec.dna)].append(rec)
            else:
                view_dict_[rec.ctx] = rec

    def populate_context_dicts(all_view_dict_, view_dict_, jtag_ctx_, dap_ctx_):
        # Helper to reduce duplication - add nodes to correct dict without help of dna
        for ctx in list(view_dict_.keys()):
            rec = view_dict_[ctx]
            if jtag_ctx_ in ctx:
                all_view_dict_[jtag_ctx].append(rec)
                del view_dict_[ctx]
            elif dap_ctx_ and dap_ctx_ in ctx:
                all_view_dict_[jtag_ctx_].append(rec)
                del view_dict_[ctx]

    # Tracking dictionaries. As we detect nodes attached to specific devices,
    # they move from the <view>_view_ctx_dicts into the all_view_dict.
    jtag_view_ctx_dict = {}
    memory_view_ctx_dict = {}
    chipscope_view_ctx_dict = {}
    debugcore_view_ctx_dict = {}
    all_view_dict: Dict = defaultdict(list)

    # Populate the all_view_dna_dict with any nodes tha that have dna
    # values. Maintain context dictionaries for fallback matching.
    if hw_server:
        populate_dna_or_context_dicts(
            scan_jtag_view(hw_server, include_dna=include_dna), all_view_dict, jtag_view_ctx_dict
        )
        populate_dna_or_context_dicts(
            scan_memory_view(hw_server, include_dna=include_dna),
            all_view_dict,
            memory_view_ctx_dict,
        )
        populate_dna_or_context_dicts(
            scan_debugcore_view(hw_server, include_dna=include_dna),
            all_view_dict,
            debugcore_view_ctx_dict,
        )
    if cs_server:
        populate_dna_or_context_dicts(
            scan_chipscope_view(cs_server, include_dna=include_dna),
            all_view_dict,
            chipscope_view_ctx_dict,
        )

    # Remaining non-dna nodes need to be matched based on their context names.
    # First match any JTAG devices and pull them out of the mix.
    # Jtag nodes give us an easy context to match other nodes with and remove
    # some from the list.

    for jtag_ctx in list(jtag_view_ctx_dict.keys()):
        jtag_rec = jtag_view_ctx_dict[jtag_ctx]
        all_view_dict[jtag_ctx].append(jtag_rec)
        del jtag_view_ctx_dict[jtag_ctx]
        populate_context_dicts(all_view_dict, memory_view_ctx_dict, jtag_ctx, jtag_rec.dap_ctx)
        populate_context_dicts(all_view_dict, debugcore_view_ctx_dict, jtag_ctx, jtag_rec.dap_ctx)
        populate_context_dicts(all_view_dict, chipscope_view_ctx_dict, jtag_ctx, jtag_rec.dap_ctx)

    # What remains are nodes not attached to a jtag node and has no dna...
    # Here we glue the remaining nodes together based on debugcore contexts.
    # Not much verification with this... If we get here, the
    # device list may end up with duplicate devices for DAP and DPC
    # because the context matching is not perfect and there is no dna key.
    # This likely happens for XVC MM testing or emulation flows where the
    # device is not fully configured in hw_server like real hardware.

    for dc_ctx in list(debugcore_view_ctx_dict.keys()):
        dc_rec = debugcore_view_ctx_dict[dc_ctx]
        all_view_dict[dc_ctx].append(dc_rec)
        del debugcore_view_ctx_dict[dc_ctx]
        for mem_ctx in list(memory_view_ctx_dict.keys()):
            if dc_ctx in mem_ctx:
                mem_rec = memory_view_ctx_dict[mem_ctx]
                all_view_dict[dc_ctx].append(mem_rec)
                del memory_view_ctx_dict[mem_ctx]
        for cs_ctx in list(chipscope_view_ctx_dict.keys()):
            if dc_ctx in cs_ctx:
                cs_rec = chipscope_view_ctx_dict[cs_ctx]
                all_view_dict[dc_ctx].append(cs_rec)
                del chipscope_view_ctx_dict[cs_ctx]
    return all_view_dict



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import json
import sys
import zipfile
from abc import abstractmethod
from collections import defaultdict
from dataclasses import dataclass, asdict
from datetime import datetime
from io import TextIOBase, BytesIO, StringIO
from itertools import islice
from pprint import pformat
from typing import Generator, Dict, List, Union, Optional
from zipfile import ZipFile

from chipscopy.api.ila import ILABitRange, ILAProbeRadix
import chipscopy
from chipscopy.shared.ila_util import bin_reversed_to_hex_values
from chipscopy.utils import Enum2StrEncoder


@dataclass
class ILAWaveformProbe:
    """Probe location in a data sample."""

    name: str
    """Name of probe."""
    map: str
    """Location string"""
    map_range: List[ILABitRange]
    """List of bit ranges. See :class:`~chipscopy.api.ila.ILABitRange`"""
    is_bus: bool
    """True for bus probes"""
    bus_left_index: Optional[int] = None
    """ Bus left index. E.g. 5 in probe ``counter[5:0]``"""
    bus_right_index: Optional[int] = None
    """Bus right index. E.g. 0 in probe ``counter[5:0]``"""
    display_radix: ILAProbeRadix = ILAProbeRadix.HEX
    """Display radix, when exporting waveform data. Default is ILAProbeRadix.HEX"""
    enum_def: Optional[enum.EnumMeta] = None
    """Enum class defining {name:int} enum values, for this probe."""

    def length(self) -> int:
        return sum(mr.length for mr in self.map_range)

    def bus_range_str(self) -> str:
        if not self.is_bus:
            res = ""
        elif self.bus_left_index == self.bus_right_index:
            res = f"[{self.bus_left_index}]"
        else:
            res = f"[{self.bus_left_index}:{self.bus_right_index}]"
        return res


@dataclass
class ILAWaveform:
    """Waveform data, with data probe information."""

    width: int
    """Sample bit width."""
    sample_count: int
    """Number of data samples."""
    trigger_position: List[int]
    """Trigger position index, for each data window."""
    window_size: int
    """Number of samples in a window."""
    probes: Dict[str, ILAWaveformProbe]
    """Dict of {probe name, waveform probe}   See :class:`ILAWaveformProbe`"""
    data: bytearray
    """
    Waveform data.

    Samples are aligned on byte boundary. 

    This formula can be used to read a bit from the data:
    ::

        bytes_per_sample = len(data) // sample_count

        def get_bit_value(data: bytearray, bytes_per_sample: int,
                          sample_index: int, data_bit_index: int) -> bool:
            byte_value = data[sample_index * bytes_per_sample + data_bit_index // 8]
            mask = 1 << (data_bit_index & 0x7)
            return (byte_value & mask) != 0

    """
    gap_index: Optional[int] = None
    """
    None or 0, if the waveform has no gaps. 
    If the value is >0, one sample bit is reserved to indicate which samples are gaps,
    i.e. the samples with unknown values. 'gap_index' gives the bit location within the sample data.
    """

    def bytes_per_sample(self) -> int:
        return len(self.data) // self.sample_count

    def get_window_count(self) -> int:
        return len(self.trigger_position)

    def set_sample(self, sample_index: int, sample: bytearray) -> None:
        """Sample may have more bytes than waveform samples have. Erase any gap bit."""
        sample_byte_count = self.bytes_per_sample()
        copy_byte_count = min(sample_byte_count, len(sample))
        start = sample_byte_count * sample_index
        self.data[start : start + copy_byte_count] = sample[0:copy_byte_count]
        if not self.gap_index:
            return
        gap_byte_index, gap_bit_index = divmod(self.width, 8)
        mask = 0xFF ^ (1 << gap_bit_index)
        self.data[start + gap_byte_index] &= mask

    def export_waveform(
        self,
        export_format: str = "CSV",
        fh_or_filepath: Union[TextIOBase, BytesIO, str] = sys.stdout,
        probe_names: Optional[List[str]] = None,
        start_window_idx: int = 0,
        window_count: Optional[int] = None,
        start_sample_idx: int = 0,
        sample_count: Optional[int] = None,
        include_gap: bool = False,
        compression: int = zipfile.ZIP_DEFLATED,
        compresslevel=None,
    ) -> None:
        """
        Export a waveform in CSV, VCD or CITF format, to a file or in-memory buffer.
        By default, all samples for all probes are exported, but it is
        possible to select which probes and window/sample ranges for CSV/VCD formats.

        ================================ ======================== ==============================
        Argument/Parameter               Type                     Supported by Export Format
        ================================ ======================== ==============================
        export_format                    str                      CSV, VCD, CITF
        fh_or_filepath                   TextIOBase               CSV, VCD
        fh_or_filepath                   BytesIO                            CITF
        fh_or_filepath                   str                      CSV, VCD, CITF
        probe_names                      List[str]                CSV, VCD
        start_window                     int                      CSV, VCD
        start_sample_idx                 int                      CSV, VCD
        sample_count                     int                      CSV, VCD
        include_gap                      bool                     CSV, VCD
        include_gap                      bool                     CSV, VCD
        compression                      int                                CITF
        compresslevel                    int                                CITF
        ================================ ======================== ==============================


        Args:
            export_format (str):  Alternatives for output format.

                - 'CSV' - Comma Separated Value Format. Default.
                - 'VCD' - Value Change Dump.
                - 'CITF' - ChipScoPy ILA Trace Format. Export of a whole ILA waveform to a compressed archive.


            fh_or_filepath (TextIOBase, BytesIO, str): File object handle or filepath string. Default is `sys.stdout`.
                If the argument is a file object, closing and opening the file is the responsibility of the caller.
                If argument is a string, the file will be opened and closed by the function.

            probe_names (Optional[List[str]]): List of probe names. Default 'None' means export all probes.
            start_window_idx (int): Starting window index. Default is first window.
            window_count (Optional[int]): Number of windows to export. Default is all windows.
            start_sample_idx (int): Starting sample within window. Default is first sample.
            sample_count (Optional[int]): Number of samples per window. Default is all samples.
            include_gap (bool):  Default is False. Include the pseudo "gap" 1-bit probe in the result.
            compression: Default is zipfile.ZIP_DEFLATED. See zipfile.ZipFile at https://docs.python.org/.
            compresslevel: See zipfile.ZipFile at https://docs.python.org/.

        """

        if export_format.upper() == "CITF":
            export_compressed_waveform(self, fh_or_filepath, compression, compresslevel)
            return

        if export_format.upper() != "VCD" and export_format.upper() != "CSV":
            raise ValueError(
                f'ILAWaveform.export() called with unknown export_format:"{export_format}"'
                "Supported export formats are VCD, CSV and CITF."
            )

        if isinstance(fh_or_filepath, str):
            with open(fh_or_filepath, "w", buffering=16384) as fh:
                export_waveform_to_stream(
                    self,
                    export_format,
                    fh,
                    probe_names,
                    start_window_idx,
                    window_count,
                    start_sample_idx,
                    sample_count,
                    include_gap,
                )
            return
        else:
            export_waveform_to_stream(
                self,
                export_format,
                fh_or_filepath,
                probe_names,
                start_window_idx,
                window_count,
                start_sample_idx,
                sample_count,
                include_gap,
            )

    @staticmethod
    def import_waveform(
        import_format: str,
        filepath_or_buffer: Union[str, BytesIO],
    ):
        """
        Create an ILAWaveform object from a ChipScoPy ILA Trace Format (CITF) compressed archive.
        The archive must contain these two files:

            - waveform.cfg, waveform and probe meta information.
            - waveform.data, binary waveform samples.

        Args:
            import_format (str): Format "CITF" is supported.
            filepath_or_buffer (str, BytesIO): Filepath string or in-memory buffer.

        Returns (ILAWaveform):
            Waveform object.

        """
        if import_format.upper() != "CITF":
            raise (
                f'import_waveform command called with import_format "{import_format}".'
                f' Only "CITF" format is supported.'
            )

        return import_compressed_waveform(filepath_or_buffer)

    def get_data(
        self,
        probe_names: Optional[List[str]] = None,
        start_window_idx: int = 0,
        window_count: Optional[int] = None,
        start_sample_idx: int = 0,
        sample_count: Optional[int] = None,
        include_trigger: bool = False,
        include_sample_info: bool = False,
        include_gap: bool = False,
    ) -> Dict[str, List[int]]:
        """
        Get probe waveform data as a list of int values for each probe.
        By default, all samples for all probes are included in return data,
        but it is possible to select which probes and window/sample ranges.

        Args:
            probe_names (Optional[List[str]]): List of probe names. Default 'None' means export all probes.
            start_window_idx (int): Starting window index. Default is first window.
            window_count (Optional[int]): Number of windows to export. Default is all windows.
            start_sample_idx (int): Starting sample within window. Default is first sample.
            sample_count (Optional[int]): Number of samples per window. Default is all samples.
            include_trigger (bool): Include pseudo probe with name '__TRIGGER' in result. Default is False.
            include_sample_info (bool):  Default is False. Include the following pseudo probes in result:

              - '__SAMPLE_INDEX' - Sample index
              - '__WINDOW_INDEX' - Window index.
              - '__WINDOW_SAMPLE_INDEX' - Sample index within window.

            include_gap (bool):  Default is False. If True, include the pseudo probe '__GAP' in result. \
                                 Value 1 for a gap sample. Value 0 for a regular sample.


        Returns (Dict[str, List[int]]):
            Ordered dict, in order:
              - '__TRIGGER', if argument **include_trigger** is True
              - '__SAMPLE_INDEX', if argument **include_sample_info** is True
              - '__WINDOW_INDEX', if argument **include_sample_info** is True
              - '__WINDOW_SAMPLE_INDEX', if argument **include_sample_info** is True
              - '__GAP', if argument **include_gap** is True
              - probe values in order of argument **probe_names**.

            Dict key: probe name. Dict value is list of int values, for a probe.

        """
        return get_waveform_data_values(
            self,
            probe_names,
            start_window_idx,
            window_count,
            start_sample_idx,
            sample_count,
            include_trigger,
            include_sample_info,
            include_gap,
        )

    def get_probe_data(
        self,
        probe_name: str,
        start_window_idx: int = 0,
        window_count: Optional[int] = None,
        start_sample_idx: int = 0,
        sample_count: Optional[int] = None,
    ) -> List[int]:
        """
        Get waveform data as a list of int values for one probe.
        By default, all samples for the probe are returned,
        It is possible to select window range and sample range.

        Args:
            probe_name (str): probe name.
            start_window_idx (int): Starting window index. Default is first window.
            window_count (Optional[int]): Number of windows to export. Default is all windows.
            start_sample_idx (int): Starting sample within window. Default is first sample.
            sample_count (Optional[int]): Number of samples per window. Default is all samples.

        Returns (List[int]):
            List probe values.

        """
        res_dict = get_waveform_data_values(
            self,
            [probe_name],
            start_window_idx,
            window_count,
            start_sample_idx,
            sample_count,
            include_trigger=False,
            include_sample_info=False,
            include_gap=False,
        )
        return res_dict[probe_name]

    def __str__(self) -> str:
        items = {key: val for key, val in self.__dict__.items() if key != "data"}
        return pformat(items, 2)

    def __repr__(self) -> str:
        data_size = len(self.data) if self.data else 0
        json_dict = {
            "width": self.width,
            "sample_count": self.sample_count,
            "trigger_position": self.trigger_position,
            "window_size": self.window_size,
            "probes": {key: asdict(val) for key, val in self.probes.items()},
            "data size": hex(data_size),
        }
        ret = json.dumps(json_dict, cls=Enum2StrEncoder, indent=4)
        return ret


def tcf_get_waveform_data(tcf_node) -> {}:
    tcf_props = tcf_node.get_property_group(["data"])
    props = {
        "width": tcf_props["trace_width"],
        "sample_count": tcf_props["trace_sample_count"],
        "trigger_position": tcf_props["trace_trigger_position"],
        "window_size": tcf_props["trace_window_size"],
        "data": tcf_props["trace_data"],
    }
    return props


class WaveformWriter(object):
    def __init__(self, file_handle: Union[TextIOBase, None], probes: [ILAWaveformProbe]):
        self._file_handle = file_handle
        self._probes = probes
        self._probe_count = len(probes)
        self._probe_names = self.make_probe_names()
        self._probe_widths = [p.length() for p in probes]

    def handle(self) -> TextIOBase:
        return self._file_handle

    def write(self, msg: str):
        self._file_handle.write(msg)

    @staticmethod
    def make_values_unknown(in_values: List[str], unknown_ch: str) -> List[str]:
        return [unknown_ch * len(val) for val in in_values]

    @abstractmethod
    def make_probe_names(self) -> [str]:
        return [p.name for p in self._probes]

    @abstractmethod
    def init(self):
        pass

    @abstractmethod
    def write_sample(
        self,
        sample_position: int,
        window_index: int,
        sample_in_window_index: int,
        is_trigger: bool,
        probe_binary_reversed_values: [str],
        is_last_sample: bool,
        is_gap: bool,
    ) -> None:
        """

        Args:
            sample_position (int):
            window_index (int):
            sample_in_window_index (int):
            is_trigger (bool):
            probe_binary_reversed_values ([str]): binary string values each start lsb at position zero.
            is_last_sample(bool): Last sample in waveform.
            is_gap (bool): True ig no data available for the sample.

        Returns:

        """
        pass


class WaveformWriterCSV(WaveformWriter):
    def __init__(self, file_handle: TextIOBase, probes: [ILAWaveformProbe], include_gap: bool):
        WaveformWriter.__init__(self, file_handle, probes)
        self._include_gap = include_gap

    def init(self):
        self.write("Sample in Buffer,Sample in Window,TRIGGER,")
        if self._include_gap:
            self.write("GAP,")
        self.write(",".join(self._probe_names))
        self.write("\nRadix - UNSIGNED,UNSIGNED,UNSIGNED,")
        # Currently, HEX is the only supported radix
        self.write(",".join(["HEX"] * self._probe_count))
        self.write("\n")

    def make_probe_names(self) -> [str]:
        return [p.name + p.bus_range_str() for p in self._probes]

    def write_sample(
        self,
        sample_position: int,
        window_index: int,
        sample_in_window_index: int,
        is_trigger: bool,
        probe_binary_reversed_values: [str],
        is_last_sample: bool,
        is_gap: bool,
    ) -> None:
        hex_values = bin_reversed_to_hex_values(probe_binary_reversed_values)
        if is_gap:
            hex_values = WaveformWriter.make_values_unknown(hex_values, "X")
        hex_values_str = ",".join(hex_values)
        trig_mark = "1" if is_trigger else "0"
        if self._include_gap:
            gap_value = "1" if is_gap else "0"
            self.write(
                f"{sample_position},{sample_in_window_index},{trig_mark},{gap_value},{hex_values_str}\n"
            )
        else:
            self.write(f"{sample_position},{sample_in_window_index},{trig_mark},{hex_values_str}\n")


class WaveformWriterToDict(WaveformWriter):
    def __init__(
        self,
        probes: [ILAWaveformProbe],
        include_trigger: bool,
        include_sample_info: bool,
        include_gap: bool,
    ):
        WaveformWriter.__init__(self, None, probes)
        self._include_trigger = include_trigger
        self._include_sample_info = include_sample_info
        self._include_gap = include_gap
        self._result = defaultdict(list)

    def init(self):
        return

    def get_result(self) -> Dict[str, List[int]]:
        return self._result

    def write_sample(
        self,
        sample_position: int,
        window_index: int,
        sample_in_window_index: int,
        is_trigger: bool,
        probe_binary_reversed_values: [str],
        is_last_sample: bool,
        is_gap: bool,
    ) -> None:
        if self._include_trigger:
            self._result["__TRIGGER"].append(1 if is_trigger else 0)
        if self._include_sample_info:
            self._result["__SAMPLE_INDEX"].append(sample_position)
            self._result["__WINDOW_INDEX"].append(window_index)
            self._result["__WINDOW_SAMPLE_INDEX"].append(sample_in_window_index)
        if self._include_gap:
            self._result["__GAP"].append(1 if is_gap else 0)

        int_values = [int(r_val[::-1], 2) for r_val in probe_binary_reversed_values]
        for probe_name, val in zip(self._probe_names, int_values):
            self._result[probe_name].append(val)


class WaveformWriterVCD(WaveformWriter):
    """Value Change Dump format. See Wikipedia and IEEE Std 1364-2001."""

    def __init__(self, file_handle: TextIOBase, probes: [ILAWaveformProbe], include_gap: bool):
        WaveformWriter.__init__(self, file_handle, probes)
        self._values = None
        self._vars: List[str] = None
        self._trigger_var = None
        self._window_var = None
        self._gap_var = None
        self._include_gap = include_gap
        self._prev_sample_is_trigger = None
        self._prev_window_index = None
        self._prev_sample_is_gap = None

    @staticmethod
    def _generate_vars() -> Generator[str, None, None]:
        """Supports up to 94 + 94*94 + 94*94*94 = 839,514 probes (or variables)."""
        for x in range(ord("!"), ord("~")):
            yield chr(x)
        for xx in range(ord("!"), ord("~")):
            for yy in range(ord("!"), ord("~")):
                yield chr(xx) + chr(yy)
        for xxx in range(ord("!"), ord("~")):
            for yyy in range(ord("!"), ord("~")):
                for zzz in range(ord("!"), ord("~")):
                    yield chr(xxx) + chr(yyy) + chr(zzz)

    def _write_variable_definitions(self):
        generate_vars = WaveformWriterVCD._generate_vars
        self._values = [None] * self._probe_count
        self._trigger_var, self._window_var, self._gap_var, *self._vars = islice(
            generate_vars(),
            self._probe_count + 3,
        )

        for width, var, probe_name in zip(self._probe_widths, self._vars, self._probe_names):
            self.write(f"$var reg {width} {var} {probe_name} $end\n")
        self.write(
            f"$var reg 1 {self._trigger_var} _TRIGGER $end\n"
            f"$var reg 1 {self._window_var} _WINDOW $end\n"
        )
        if self._include_gap:
            self.write(f"$var reg 1 {self._gap_var} _GAP $end\n")

    def make_probe_names(self) -> [str]:
        return [p.name + " " + p.bus_range_str() for p in self._probes]

    def init(self):
        now = "{:%Y-%m-%d %H:%M:%S}".format(datetime.now())
        # adjust for trigger position
        hdr_1 = f"$date\n        {now}\n$end\n"

        hdr_2 = (
            "$version\n"
            + f"        ChipScoPy Version {chipscopy.__version__}\n"
            + "$end\n"
            + "$timescale\n        1ps\n$end\n"
            + "$scope module dut $end\n"
        )

        hdr_3 = "$upscope $end\n$enddefinitions $end\n"

        self.write(hdr_1 + hdr_2)
        self._write_variable_definitions()
        self.write(hdr_3)

    def write_sample(
        self,
        sample_position: int,
        window_index,
        sample_in_window_index,
        is_trigger,
        probe_binary_reversed_values: [str],
        is_last_sample,
        is_gap: bool,
    ):
        time_written = False

        def write_value(var: str, reversed_value: str) -> None:
            nonlocal time_written
            if not time_written:
                self.write(f"#{sample_position}\n")
                time_written = True

            if len(reversed_value) == 1:
                self.write(f"{reversed_value}{var}\n")
                return

            # Remove leading zeroes
            if is_gap and reversed_value[0] == "x":
                value = reversed_value
            else:
                # Remove leading zeroes
                msb_idx = reversed_value.rfind("1")
                if msb_idx < 0:
                    msb_idx = 0
                value = reversed_value[msb_idx::-1]
            self.write(f"b{value} {var}\n")

        # Write time, for last sample, even if no changes.
        if is_last_sample:
            self.write(f"#{sample_position}\n")
            time_written = True

        # Trigger value
        if is_trigger != self._prev_sample_is_trigger:
            write_value(self._trigger_var, "1" if is_trigger else "0")
            self._prev_sample_is_trigger = is_trigger

        # Window marker.
        if window_index != self._prev_window_index:
            write_value(self._window_var, "1" if window_index % 2 else "0")
            self._prev_window_index = window_index

        # gap value
        if self._include_gap and is_gap != self._prev_sample_is_gap:
            write_value(self._gap_var, "1" if is_gap else "0")
            self._prev_sample_is_gap = is_gap

        # Regular values
        if is_gap:
            probe_binary_reversed_values = WaveformWriter.make_values_unknown(
                probe_binary_reversed_values, "x"
            )
        for idx, new_val in enumerate(probe_binary_reversed_values):
            if new_val == self._values[idx]:
                continue

            self._values[idx] = new_val
            write_value(self._vars[idx], new_val)


def export_waveform_to_stream(
    waveform: ILAWaveform,
    export_format: str,
    stream_handle: TextIOBase,
    probe_names: [str],
    start_window_idx: int,
    window_count: Optional[int],
    start_sample_idx: int,
    sample_count: Optional[int],
    include_gap: bool,
) -> None:
    """Arguments documented in calling API function"""
    if probe_names:
        probes = [waveform.probes.get(p_name, None) for p_name in probe_names]
        if not all(probes):
            bad_names = set(probe_names) - set(waveform.probes.keys())
            raise KeyError(f"export_waveform() called with non-existent probe_name:\n  {bad_names}")
    else:
        probes = waveform.probes.values()

    if export_format.upper() == "CSV":
        waveform_writer = WaveformWriterCSV(stream_handle, probes, include_gap)
    elif export_format.upper() == "VCD":
        waveform_writer = WaveformWriterVCD(stream_handle, probes, include_gap)
    else:
        raise ValueError(
            f'ILAWaveform.export_waveform() called with non-supported format "{export_format}"'
        )

    _export_waveform(
        waveform,
        waveform_writer,
        probes,
        start_window_idx,
        window_count,
        start_sample_idx,
        sample_count,
        "export_waveform()",
    )


def get_waveform_data_values(
    waveform: ILAWaveform,
    probe_names: [str],
    start_window_idx: int,
    window_count: Optional[int],
    start_sample_idx: int,
    sample_count: Optional[int],
    include_trigger: bool,
    include_sample_info: bool,
    include_gap: bool,
) -> Dict[str, List[int]]:
    """Arguments documented in calling API function"""
    if probe_names:
        probes = [waveform.probes.get(p_name, None) for p_name in probe_names]
        if not all(probes):
            bad_names = set(probe_names) - set(waveform.probes.keys())
            raise KeyError(
                f"ILAWaveform.get_data() called with non-existent probe name(s):\n  {bad_names}"
            )
    else:
        probes = waveform.probes.values()

    waveform_writer = WaveformWriterToDict(
        probes, include_trigger, include_sample_info, include_gap
    )
    _export_waveform(
        waveform,
        waveform_writer,
        probes,
        start_window_idx,
        window_count,
        start_sample_idx,
        sample_count,
        "ILAWaveform.get_data()",
    )

    return waveform_writer.get_result()


def _export_waveform(
    waveform: ILAWaveform,
    writer: WaveformWriter,
    probes: [ILAWaveformProbe],
    start_window_idx: int,
    window_count: Optional[int],
    start_sample_idx: int,
    sample_count: Optional[int],
    calling_function: str,
) -> None:
    def get_sample_binary_values(sample_data: memoryview) -> [str]:
        sample_int_value = int.from_bytes(sample_data, byteorder="little", signed=False)
        reversed_bin_value = f"{sample_int_value:0{waveform.width}b}"[::-1]
        values = []
        for probe in probes:
            # Each probe value can be made up of multiple bit range slices.
            slices = [reversed_bin_value[br.index : br.index + br.length] for br in probe.map_range]
            p_val = "".join(slices)
            values.append(p_val)
        return values

    def is_gap(sample_high_byte: int, gap_index: Optional[int]) -> bool:
        if gap_index and sample_high_byte & (1 << (gap_index % 8)):
            return True
        return False

    if not window_count:
        window_count = waveform.get_window_count()
    if not sample_count:
        sample_count = waveform.window_size

    w_size = waveform.window_size
    max_window_count = (waveform.sample_count + w_size - 1) // w_size
    if start_window_idx < 0 or start_window_idx >= max_window_count:
        raise ValueError(
            f'{calling_function} function argument start_window="{start_window_idx}" '
            f"must be in the range [0-{max_window_count - 1}]"
        )
    if start_sample_idx < 0 or start_sample_idx >= w_size:
        raise ValueError(
            f'{calling_function} function argument "start_sample="{start_sample_idx}" '
            f"must be in the range [0-{w_size - 1}]"
        )
    if sample_count < 1 or sample_count > w_size - start_sample_idx:
        raise ValueError(
            f'{calling_function} function argument "sample_count="{sample_count}" '
            f"must be in the range [1-{w_size - start_sample_idx}], "
            f'since start_sample_idx="{start_sample_idx}".'
        )
    if window_count < 1 or window_count > max_window_count - start_window_idx:
        raise ValueError(
            f'{calling_function} function argument "window_count="{window_count}" '
            f"must be in the range [1-{max_window_count - start_window_idx}], "
            f'since start_window_idx="{start_window_idx}".'
        )

    sample_byte_count = waveform.bytes_per_sample()
    raw_samples = memoryview(waveform.data)
    sample_is_gap = False
    writer.init()

    for window_idx in range(start_window_idx, start_window_idx + window_count):
        window_start_sample_idx = window_idx * w_size
        trigger_sample_idx = waveform.trigger_position[window_idx]
        for sample_idx in range(
            window_start_sample_idx + start_sample_idx,
            window_start_sample_idx + start_sample_idx + sample_count,
        ):
            if sample_idx >= waveform.sample_count:
                # last window is not full.
                break
            raw_sample_idx = sample_idx * sample_byte_count
            raw_sample_idx_end = raw_sample_idx + sample_byte_count
            sample_idx_in_window = sample_idx - window_start_sample_idx
            bin_values = get_sample_binary_values(raw_samples[raw_sample_idx:raw_sample_idx_end])
            if waveform.gap_index:
                sample_is_gap = is_gap(raw_samples[raw_sample_idx_end - 1], waveform.gap_index)
            writer.write_sample(
                sample_idx,
                window_idx,
                sample_idx_in_window,
                sample_idx_in_window == trigger_sample_idx,
                bin_values,
                sample_idx + 1 == waveform.sample_count,
                sample_is_gap,
            )


WAVEFORM_ARCHIVE_VERSION = 1


class Waveform2StrEncoder(json.JSONEncoder):
    @staticmethod
    def encode_map_range(obj):
        # Convert ILABitRange tuple to a dict, e.g. [{"index": 0, "length": 2}]
        if isinstance(obj, list):
            return [elem._asdict() if isinstance(elem, ILABitRange) else elem for elem in obj]
        return obj

    def default(self, obj):
        if isinstance(obj, ILAWaveformProbe):
            return {
                key: Waveform2StrEncoder.encode_map_range(val) for key, val in asdict(obj).items()
            }
        if isinstance(obj, ILABitRange):
            return obj._asdict()
        if isinstance(obj, ILAProbeRadix):
            return str(obj.name)
        if isinstance(obj, enum.EnumMeta):
            # Encode as 2-item list: [<name>, <dict of enum value items>]
            items = {item.name: item.value for item in list(obj)}
            return [obj.__name__, items]
        return json.JSONEncoder.default(self, obj)


def export_compressed_waveform(
    waveform: ILAWaveform,
    filepath_or_buffer: Union[str, BytesIO],
    compression: int,
    compresslevel: int,
) -> None:
    waveform_dict = {
        "version": 1,
        "gap_index": waveform.gap_index,
        "probes": waveform.probes,
        "sample_count": waveform.sample_count,
        "trigger_position": waveform.trigger_position,
        "window_size": waveform.window_size,
        "width": waveform.width,
    }

    json_str = json.dumps(waveform_dict, cls=Waveform2StrEncoder, indent=4)

    with ZipFile(
        filepath_or_buffer,
        mode="w",
        allowZip64=True,
        compression=compression,
        compresslevel=compresslevel,
    ) as zf:
        zf.writestr("waveform.cfg", json_str)
        zf.writestr("waveform.data", waveform.data)


def decode_waveform_from_json(json_str: str, in_data: bytearray) -> ILAWaveform:
    def decode_map_range(in_range: List[Dict[str, int]]) -> List[ILABitRange]:
        res = [ILABitRange(**dd) for dd in in_range]
        return res

    def decode_probe(probe_name: str, in_dict: Dict) -> ILAWaveformProbe:
        pd = {}
        for key, value in in_dict.items():
            try:
                if key == "map_range":
                    pd[key] = decode_map_range(value)
                elif key == "display_radix":
                    pd[key] = ILAProbeRadix[value]
                elif key == "enum_def":
                    if isinstance(value, list) and len(value) == 2:
                        pd[key] = enum.Enum(value[0], value[1])
                    else:
                        pd[key] = None
                else:
                    pd[key] = value
            except Exception as ex:
                raise ValueError(f'Error reading ILA waveform probe "{probe_name}"') from ex

        res = ILAWaveformProbe(**pd)
        return res

    def decode_probes(in_probes: Dict) -> Dict[str, ILAWaveformProbe]:
        res = {name: decode_probe(name, probe_dict) for name, probe_dict in in_probes.items()}
        return res

    # Decode waveform json string.
    json_dict = json.load(StringIO(json_str))
    wd = dict()
    wd["data"] = in_data
    wd["probes"] = decode_probes(json_dict.get("probes", dict()))
    for key, value in json_dict.items():
        if key in ["data", "probes"]:
            # data and probes, handled above.
            pass
        elif key == "version":
            if value > WAVEFORM_ARCHIVE_VERSION:
                raise ValueError(
                    f'waveform archive version "{value}" is not supported.'
                    f'Only versions "<={WAVEFORM_ARCHIVE_VERSION}" are supported.'
                )
        else:
            wd[key] = value

    return ILAWaveform(**wd)


def import_compressed_waveform(filepath_or_buffer: Union[str, BytesIO]) -> ILAWaveform:
    with ZipFile(filepath_or_buffer, mode="r") as zf:
        json_str = zf.read("waveform.cfg").decode()
        data = bytearray(zf.read("waveform.data"))

    waveform = decode_waveform_from_json(json_str, data)
    return waveform



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2024, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Any, NewType
from chipscopy.tcf import channel, errors
from chipscopy.tcf.services import DoneHWCommand, memory

MemoryNode = NewType("MemoryNode", Dict[str, Any])

Memory_SERVICE = "Memory"


class Range(object):
    offs = 0
    size = 0
    stat = 0
    msg = None

    def __lt__(self, o):
        return self.__cmp__(o) == -1

    def __cmp__(self, o):
        if self.offs < o.offs:
            return -1
        if self.offs > o.offs:
            return +1
        return 0


class MemoryErrorReport(errors.ErrorReport, memory.MemoryError, memory.ErrorOffset):
    def __init__(self, msg, attrs, addr, ranges):
        super(MemoryErrorReport, self).__init__(msg, attrs)
        if ranges is None:
            self.ranges = None
        else:
            self.ranges = []
            for m in ranges:
                r = Range()
                x = m.get(memory.ErrorOffset.RANGE_KEY_ADDR)
                if isinstance(x, str):
                    y = int(x)
                else:
                    y = x
                r.offs = y - addr
                r.size = m.get(memory.ErrorOffset.RANGE_KEY_SIZE)
                r.stat = m.get(memory.ErrorOffset.RANGE_KEY_STAT)
                key = memory.ErrorOffset.RANGE_KEY_MSG
                r.msg = errors.toErrorString(m.get(key))
                assert r.offs >= 0
                assert r.size >= 0
                self.ranges.append(r)
            self.ranges.sort()

    def getMessage(self, offset):
        if self.ranges is None:
            return None
        l = 0
        h = len(self.ranges) - 1
        while l <= h:
            n = int((l + h) / 2)
            r = self.ranges[n]
            if r.offs > offset:
                h = n - 1
            elif offset >= r.offs + r.size:
                l = n + 1
            else:
                return r.msg
        return None

    def getStatus(self, offset):
        if self.ranges is None:
            return memory.ErrorOffset.BYTE_UNKNOWN
        l = 0
        h = len(self.ranges) - 1
        while l <= h:
            n = int((l + h) / 2)
            r = self.ranges[n]
            if r.offs > offset:
                h = n - 1
            elif offset >= r.offs + r.size:
                l = n + 1
            else:
                return r.stat
        return memory.ErrorOffset.BYTE_UNKNOWN


def toMemoryError(service, cmd, addr, data, ranges):
    if data is None:
        return None
    code = data.get(errors.ERROR_CODE)
    if len(cmd) > 72:
        cmd = cmd[0:72] + "..."
    msg = "TCF command exception:\nCommand: %s\n" % cmd + "Exception: %s\nError code: %d" % (
        errors.toErrorString(data),
        code,
    )
    e = MemoryErrorReport(msg, data, addr, ranges)
    return e


class MemContext(memory.MemoryContext):
    def __init__(self, service, props):
        super(MemContext, self).__init__(props)
        self.service = service

    def fill(self, addr, word_size, value, size, mode, done):
        service = self.service
        context_id = self.getID()
        done = service._makeCallback(done)
        cmd = "fill"

        def done_hw(self, token, error, args):
            e = None
            if error:
                # XXX : fle : Exception.message does not exist in python3,
                #             better use str(Exception)
                e = error
            else:
                assert len(args) == 2
                e = toMemoryError(service.getName(), cmd, addr, args[0], args[1])
            done.doneMemory(token, e)

        return service.send_old_command(
            cmd, (context_id, addr, word_size, size, mode, value), done_hw
        )

    def get(self, addr, word_size, buf, offs, size, mode, done):
        service = self.service
        context_id = self.getID()
        done = service._makeCallback(done)
        cmd = "get"

        def done_hw(token, error, args):
            e = None
            if error:
                # Exception.message does not exist in python3, better use
                # str(Exception)
                e = memory.MemoryError(str(error))
            else:
                assert len(args) == 3
                byts = channel.toByteArray(args[0])
                assert len(byts) <= size
                buf[offs : offs + len(byts)] = byts
                e = toMemoryError(service.getName(), cmd, addr, args[1], args[2])
            done.doneMemory(token, e)

        return service.send_old_command(cmd, (context_id, addr, word_size, size, mode), done_hw)

    def set(self, addr, word_size, buf, offs, size, mode, done):
        service = self.service
        context_id = self.getID()
        done = service._makeCallback(done)
        cmd = "set"

        def done_hw(token, error, args):
            e = None
            if error:
                # XXX : fle : Exception.message does not exist in python3,
                #             better use str(Exception)
                e = memory.MemoryError(str(error))
            else:
                assert len(args) == 2
                e = toMemoryError(service.getName(), cmd, addr, args[0], args[1])
            done.doneMemory(token, e)

        return service.send_old_command(
            cmd,
            (context_id, addr, word_size, size, mode, bytearray(buf[offs : offs + size])),
            done_hw,
        )


class MemoryProxy(memory.MemoryService):
    """TCF Memory service interface."""

    def __init__(self, channel):
        super(MemoryProxy, self).__init__(channel)
        self.listeners = {}

    def getName(self):
        """
        Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return Memory_SERVICE

    def get_context(self, context_id: str, done: DoneHWCommand) -> Dict[str, Any]:
        """
        Gets context information on a given context id

        :param context_id: Context id of the context
        :param done: Callback with result and any error.
        :return: A dictionary of context parameters
        """
        done = self._makeCallback(done)
        service = self

        class DoneContext(DoneHWCommand):
            def doneHW(self, token, error, args):
                ctx = None
                if not error:
                    assert len(args) == 1
                    if args[0]:
                        ctx = MemContext(service, args[0])
                done.doneGetContext(token, error, ctx)

        return self.send_command("getContext", (context_id,), DoneContext())

    def get_children(self, context_id: str, done: DoneHWCommand) -> List[str]:
        """
        Gets the children of a given context.  If no context is given then gets top level contexts.

        :param context_id: Context id of desired context
        :param done: Callback with result and any error.
        :return: List of child contexts
        """
        done = self._makeCallback(done)
        if not context_id:
            context_id = ""

        class DoneChildren(DoneHWCommand):
            def doneHW(self, token, error, args):
                contexts = None
                if not error:
                    assert len(args) == 1
                    contexts = args[0]
                done.doneGetChildren(token, error, contexts)

        return self.send_command("getChildren", (context_id,), DoneChildren())

    def add_listener(self, listener):
        """Add Memory service event listener.
        :param listener: Event listener implementation.
        """
        l = ChannelEventListener(self, listener)
        self.channel.addEventListener(self, l)
        self.listeners[listener] = l

    def remove_listener(self, listener):
        """Remove Memory service event listener.
        :param listener: Event listener implementation.
        """
        l = self.listeners.get(listener)
        if l:
            del self.listeners[listener]
            self.channel.removeEventListener(self, l)

    getContext = get_context
    getChildren = get_children
    addListener = add_listener
    removeListener = remove_listener


class ChannelEventListener(channel.EventListener):
    def __init__(self, service, listener):
        self.service = service
        self.listener = listener
        self._event_handlers = {
            "contextAdded": self._context_added,
            "contextChanged": self._context_changed,
            "contextRemoved": self._context_removed,
            "memoryChanged": self._memory_changed,
        }

    def _context_added(self, args):
        assert len(args) == 1
        self.listener.contextAdded(_toContextArray(self.service, args[0]))

    def _context_changed(self, args):
        assert len(args) == 1
        self.listener.contextChanged(_toContextArray(self.service, args[0]))

    def _context_removed(self, args):
        assert len(args) == 1
        self.listener.contextRemoved(args[0])

    def _memory_changed(self, args):
        assert len(args) == 2
        self.listener.memoryChanged(args[0], _toAddrArray(args[1]), _toSizeArray(args[1]))


def _toContextArray(svc, o):
    if o is None:
        return None
    ctx = []
    for m in o:
        ctx.append(MemContext(svc, m))
    return ctx


def _toSizeArray(o):
    if o is None:
        return None
    return [m.get("size", 0) for m in o]


def _toAddrArray(o):
    if o is None:
        return None
    return [m.get("addr") for m in o]



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2024, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Any, NewType
from chipscopy.tcf import channel
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.tcf.services import runcontrol

RunControlNode = NewType("RunControlNode", Dict[str, Any])

RunControl_SERVICE = "RunControl"


class RunContext(runcontrol.RunControlContext):
    def __init__(self, service, props):
        super(RunContext, self).__init__(props)
        self.service = service

    def getISA(self, address, done):
        service = self.service
        context_id = self.getID()
        done = service._makeCallback(done)
        cmd = "getISA"

        def done_hw(token, error, args):
            isa = None
            if not error:
                assert len(args) == 2
                error = args[0]
                isa = runcontrol.RunControlISA(args[1])
            done.doneGetISA(token, error, isa)

        return service.send_old_command(cmd, (context_id, address), done_hw)

    def getState(self, done):
        service = self.service
        context_id = self.getID()
        done = service._makeCallback(done)
        cmd = "getState"

        def done_hw(token, error, args):
            susp = False
            pc = None
            reason = None
            states = None
            if not error:
                assert len(args) == 5
                error = args[0]
                susp = args[1]
                if args[2] is not None:
                    pc = int(args[2])
                reason = args[3]
                states = args[4]
            done.doneGetState(token, error, susp, pc, reason, states)

        return service.send_old_command(cmd, (context_id,), done_hw)

    def detach(self, done):
        return self._command("detach", (self.getID(),), done)

    def resume(self, mode, count, params, done):
        if not params:
            return self._command("resume", (self.getID(), mode, count), done)
        else:
            return self._command("resume", (self.getID(), mode, count, params), done)

    def suspend(self, done):
        return self._command("suspend", (self.getID(),), done)

    def terminate(self, done):
        return self._command("terminate", (self.getID(),), done)

    def _command(self, cmd, args, done):
        service = self.service
        done = service._makeCallback(done)

        def done_hw(token, error, args):
            if not error:
                assert len(args) == 1
                error = args[0]
            done.doneCommand(token, error)

        return service.send_old_command(cmd, args, done_hw)


class RunControlProxy(runcontrol.RunControlService):
    """TCF RunControl service interface."""

    def __init__(self, channel):
        super(RunControlProxy, self).__init__(channel)
        self.listeners = {}

    def getName(self):
        """
        Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return RunControl_SERVICE

    def get_context(self, context_id: str, done: DoneHWCommand) -> Dict[str, Any]:
        """
        Gets context information on a given context id

        :param context_id: Context id of the context
        :param done: Callback with result and any error.
        :return: A dictionary of context parameters
        """
        done = self._makeCallback(done)
        service = self

        class DoneContext(DoneHWCommand):
            def doneHW(self, token, error, args):
                ctx = None
                if not error:
                    assert len(args) == 1
                    if args[0]:
                        ctx = RunContext(service, args[0])
                done.doneGetContext(token, error, ctx)

        return self.send_command("getContext", (context_id,), DoneContext())

    def get_children(self, context_id: str, done: DoneHWCommand) -> List[str]:
        """
        Gets the children of a given context.  If no context is given then gets top level contexts.

        :param context_id: Context id of desired context
        :param done: Callback with result and any error.
        :return: List of child contexts
        """
        done = self._makeCallback(done)
        if not context_id:
            context_id = ""

        class DoneChildren(DoneHWCommand):
            def doneHW(self, token, error, args):
                contexts = None
                if not error:
                    assert len(args) == 1
                    contexts = args[0]
                done.doneGetChildren(token, error, contexts)

        return self.send_command("getChildren", (context_id,), DoneChildren())

    def add_listener(self, listener):
        """Add RunControl service event listener.
        :param listener: Event listener implementation.
        """
        l = ChannelEventListener(self, listener)
        self.channel.addEventListener(self, l)
        self.listeners[listener] = l

    def remove_listener(self, listener):
        """Remove RunControl service event listener.
        :param listener: Event listener implementation.
        """
        l = self.listeners.get(listener)
        if l:
            del self.listeners[listener]
            self.channel.removeEventListener(self, l)

    getContext = get_context
    getChildren = get_children
    addListener = add_listener
    removeListener = remove_listener


class ChannelEventListener(channel.EventListener):
    def __init__(self, service, listener):
        self.service = service
        self.listener = listener
        self._event_handlers = {
            "contextSuspended": self._context_suspended,
            "contextResumed": self._context_resumed,
            "contextAdded": self._context_added,
            "contextChanged": self._context_changed,
            "contextRemoved": self._context_removed,
            "contextException": self._context_exception,
            "containerSuspended": self._container_suspended,
            "containerResumed": self._container_resumed,
            "contextStateChanged": self._context_state_changed,
        }

    def _context_suspended(self, args):
        assert len(args) == 4
        self.listener.contextSuspended(args[0], args[1], args[2], args[3])

    def _context_resumed(self, args):
        assert len(args) == 1
        self.listener.contextResumed(args[0])

    def _context_added(self, args):
        assert len(args) == 1
        self.listener.contextAdded(_toContextArray(self.service, args[0]))

    def _context_changed(self, args):
        assert len(args) == 1
        self.listener.contextChanged(_toContextArray(self.service, args[0]))

    def _context_removed(self, args):
        assert len(args) == 1
        self.listener.contextRemoved(args[0])

    def _context_exception(self, args):
        assert len(args) == 2
        self.listener.contextException(args[0], args[1])

    def _container_suspended(self, args):
        assert len(args) == 5
        self.listener.containerSuspended(args[0], args[1], args[2], args[3], args[4])

    def _container_resumed(self, args):
        assert len(args) == 1
        self.listener.containerResumed(args[0])

    def _context_state_changed(self, args):
        assert len(args) == 1
        self.listener.contextStateChanged(args[0])


def _toContextArray(svc, o):
    if o is None:
        return None
    ctx = []
    for m in o:
        ctx.append(RunContext(svc, m))
    return ctx


def _toSizeArray(o):
    if o is None:
        return None
    return [m.get("size", 0) for m in o]


def _toAddrArray(o):
    if o is None:
        return None
    return [m.get("addr") for m in o]



# *****************************************************************************
# * Modifications Copyright (C) 2021-2022, Xilinx, Inc., All rights reserved.
# * Modifications Copyright (C) 2022-2024, Advanced Micro Devices, Inc., All rights reserved..
# *
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************
try:
    import pytcf
    from ..native.services import *
    from .arguments import *
except:
    from typing import NewType
    import threading
    import inspect
    import importlib
    from .. import protocol
    from ..channel import *
    from ..channel.Command import Command
    from .arguments import *

    _providers = []
    _lock = threading.RLock()

    TokenType = NewType("TokenType", Token)


    class TCFError(Exception):
        pass


    class DoneHWCommand(object):
        """Client call back interface for generic commands."""
        def doneHW(self, token, error, args):
            """Called when memory operation command command is done.
            :param token: Command handle.
            :param error: Error object or **None**.
            """
            pass

        def __call__(self, token, error, args):
            return self.doneHW(token, error, args)


    class ServiceProvider(object):
        """Clients can implement this abstract class if they want to provide
        implementation of a local service or remote service proxy.
        """

        def get_local_service(self, channel):
            pass

        def get_service_proxy(self, channel, service_name):
            pass


    def add_service_provider(provider):
        with _lock:
            _providers.append(provider)


    def removeServiceProvider(provider):
        with _lock:
            _providers.remove(provider)


    def onChannelCreated(channel, services_by_name):
        with _lock:
            # TODO ZeroCopy support is incomplete
            # zero_copy = ZeroCopy()
            # services_by_name[zero_copy.getName()] = zero_copy
            for provider in _providers:
                try:
                    arr = provider.get_local_service(channel)
                    if not arr:
                        continue
                    for service in arr:
                        if service.getName() in services_by_name:
                            continue
                        services_by_name[service.getName()] = service
                except Exception as x:
                    protocol.log("Error calling TCF service provider", x)


    def onChannelOpened(channel, service_names, services_by_name):
        with _lock:
            for name in service_names:
                for provider in _providers:
                    try:
                        service = provider.get_service_proxy(channel, name)
                        if not service:
                            continue
                        services_by_name[name] = service
                        break
                    except Exception as x:
                        protocol.log("Error calling TCF service provider", x)
                if name in services_by_name:
                    continue
                services_by_name[name] = GenericProxy(channel, name)


    def getServiceManagerID():
        # In current implementation ServiceManager is a singleton,
        # so its ID is same as agent ID.
        return protocol.getAgentID()


    class GenericCallback(object):

        def __init__(self, callback):
            self.callback = callback

        def __getattr__(self, attr):
            if attr.startswith("done"):
                return self.callback


    class Service(object):
        """TCF service base class."""
        def getName(self):
            """Abstract method to get the service name.

            :returns: This service name
            """
            raise NotImplementedError("Abstract method")

        def __init__(self, channel):
            self.channel = channel

        def __str__(self):
            """TCF service string representation.

            :returns: The name of the service.
            """
            return self.getName()

        def _makeCallback(self, done):
            """Turn *done* into a callable.

            If *done* is already a :class:`collections.Callable`, it is returned
            as is, else, it is made callable, and returned.

            :param done: The item to make callable.

            :returns: The callable value of *done*
            """
            if inspect.isfunction(done) or inspect.ismethod(done):
                return GenericCallback(done)
            return done

        def send_command(self, name, args, done):
            done = self._makeCallback(done)
            service = self

            class HWCommand(Command):
                def __init__(self):
                    super(HWCommand, self).__init__(service.channel,
                                                    service.getName(),
                                                    name,
                                                    args)

                def done(self, error, results):
                    if not error:
                        assert len(results) > 0
                        #error = self.toError(results[0])
                        error = results[0]
                        results = results[1:]
                    done.doneHW(self.token, error, results)
            return HWCommand().token

        def send_xicom_command(self, name, args, done, progress=None):
            done = self._makeCallback(done)
            service = self

            class XicomCommand(Command):
                def __init__(self):
                    super(XicomCommand, self).__init__(service.channel,
                                                    service.getName(),
                                                    name,
                                                    to_xargs(args))

                def done(self, error, results):
                    if not error:
                        assert len(results) >= 1
                        error = results[0]
                        if isinstance(error, dict) and 'Format' in error:
                            e = Exception(error['Format'])
                            try:
                                mod = importlib.import_module(error['Module'])
                                cls = getattr(mod, error["Class"])
                                e = cls(f"TCF Error: {service.getName()} {name}: {error['Format']}")
                            except Exception:
                                pass
                            error = e
                        results = from_xargs(results[1:]) if len(results) > 1 else None
                        if results and len(results) == 1:
                            results = results[0]
                    if done:
                        done.doneHW(self.token, error, results)

                def progress_update(self, error, results):
                    if not error:
                        assert len(results) >= 1
                        error = results[0]
                        if isinstance(error, dict) and 'Format' in error:
                            e = Exception(error['Format'])
                            try:
                                mod = importlib.import_module(error['Module'])
                                cls = getattr(mod, error["Class"])
                                e = cls(f"TCF Error: {service.getName()} {name}: {error['Format']}")
                            except Exception:
                                pass
                            error = e
                        results = from_xargs(results[1:]) if len(results) > 1 else None
                        if results and len(results) == 1:
                            results = results[0]
                    if progress:
                        progress(self.token, error, results)
            return XicomCommand().token

        def send_old_xicom_command(self, name, args, done):
            done = self._makeCallback(done)
            service = self

            class OldXicomCommand(Command):
                def __init__(self):
                    super().__init__(service.channel,
                                     service.getName(),
                                     name,
                                     to_xargs(args))

                @staticmethod
                def preprocess_data(data):
                    def next_bytes(data):
                        start = data.find(b'(')
                        if start > 0 and data[start-1] == b'"':
                            start = -1
                        return start
                    data_bytes = bytearray()
                    processed_data = bytearray()
                    while len(data):
                        data_tail = b''
                        start = next_bytes(data)
                        while start >= 0:
                            count_stop = data.find(b')',start)
                            count = int(data[start+1:count_stop])
                            stop = count_stop + 1 + count
                            data_bytes += data[count_stop + 1:stop]
                            processed_data += data[:start]
                            processed_data += f"\"({count})\"".encode('utf-8')
                            data = data[stop:]
                            start = next_bytes(data)
                            arg_end = data.find(b'\x00')
                            if 0 <= arg_end < start:
                                data_tail = data[:arg_end]
                                data = data[arg_end+1:]
                                break
                            data_tail = data
                        if data_tail:
                            processed_data += data_tail
                        else:
                            processed_data += data
                            break
                        if data_bytes:
                            processed_data += b'\x00'
                            processed_data += f"({len(data_bytes)})".encode('utf-8') + data_bytes
                        if start < 0:
                            break
                        processed_data += b'\x00'
                        data_bytes.clear()
                    return processed_data

                def result(self, token, data):
                    data = OldXicomCommand.preprocess_data(data)
                    super().result(token, data)

                def done(self, error, results):
                    if not error:
                        assert len(results) >= 1
                        error = results[0]
                        if isinstance(error, dict) and 'Format' in error:
                            e = Exception(error['Format'])
                            try:
                                mod = importlib.import_module(error['Module'])
                                cls = getattr(mod, error["Class"])
                                e = cls(f"TCF Error: {service.getName()} {name}: {error['Format']}")
                            except Exception:
                                pass
                            error = e
                        results = from_xargs(results[1:]) if len(results) > 1 else None
                        if results and len(results) == 1:
                            results = results[0]
                    if done:
                        done.doneHW(self.token, error, results)
            return OldXicomCommand().token

        def send_old_command(self, name, args, done):
            done = self._makeCallback(done)
            service = self

            class OldCommand(Command):
                def __init__(self):
                    super().__init__(service.channel,
                                     service.getName(),
                                     name,
                                     to_xargs(args))

                def done(self, error, results):
                    if not error:
                        results = from_xargs(results)
                        if results and len(results) == 1:
                            results = results[0]
                    if done:
                        done.doneHW(self.token, error, results)
            return OldCommand().token


    class ServiceSync(object):
        def __init__(self, channel, service_name, raise_on_error=False):
            self.service_name = service_name
            self.chan = channel
            self.result = None
            self.raise_on_error = raise_on_error

            #services = protocol.invokeAndWait(self.chan.getRemoteServices)
            # if service_name not in services:
            #     raise Exception('No service {} found'.format(service_name))

        def __getattr__(self, attr):
            return CommandWrapper(self, attr)

        def _doneCommand(self, token, error, args):
            self.result = CommandResult(token, error, args)

        def invoke(self, command, *args, **kwargs):
            cond = threading.Condition()
            self.result = None

            def callCommand(command, *args, **kwargs):
                serv = self.chan.getRemoteService(self.service_name)

                class DoneCommand(DoneHWCommand):
                    def __init__(self, service):
                        self.service = service

                    def doneHW(self, token, error, args):
                        self.service._doneCommand(token, error, args)
                        with cond:
                            cond.notify()

                try:
                    kwargs['done'] = DoneCommand(self)
                    f = getattr(serv, command, None)
                    f(*args, **kwargs)
                except TypeError as e:
                    error = '{} service {} error: {}'.format(serv,command,str(e))
                    self._doneCommand(Token(),error,[])
                    with cond:
                        cond.notify()

            if protocol.isDispatchThread():
                callCommand(command, *args, **kwargs)
            else:
                with cond:
                    protocol.invokeLater(callCommand, command, *args, **kwargs)
                    # When inside the debugger wait longer, don't time out.
                    # todo: move this feature to the command line???
                    cond.wait(None if __debug__ else 60)

            if self.raise_on_error and self.result and self.result.error:
                if isinstance(self.result.error, Exception):
                    raise self.result.error
                else:
                    raise Exception(self.service_name + ':' + command + ': ' + str(self.result.error))
            return self.result


    class CommandResult(object):
        def __init__(self, token, error, args):
            self.token = token
            self.error = error
            # unwrap result if only one element
            # if args and len(args) == 1 and \
            #         not isinstance(args, dict):
            #     args = args[0]
            self.args = args

        def __str__(self):
            if self.error:
                return "[%s] error: %s" % (self.token.id, self.error)
            return "[%s] result: %s" % (self.token.id, self.args)
        __repr__ = __str__

        def __iter__(self):
            yield self.error
            yield self.args

        def get(self, key=None):
            if key:
                return self.args.get(key)
            return self.args


    class CommandWrapper(object):
        def __init__(self, control, command):
            self._control = control
            self._command = command

        def __call__(self, *args, **kwargs):
            return self._control.invoke(self._command, *args, **kwargs)


    class ZeroCopy(Service):
        def getName(self):
            return "ZeroCopy"


    class GenericProxy(Service):
        """Objects of GenericProxy class represent remote services, which don't
        have a proxy class defined for them.

        Clients still can use such services, but framework will not provide
        service specific utility methods for message formatting and parsing.
        """

        def __init__(self, channel, name):
            self.__channel = channel
            self.name = name

        def getName(self):
            return self.name

        def getChannel(self):
            return self.__channel


    class DefaultServiceProvider(ServiceProvider):
        package_base = str(__package__) + ".remote"

        def get_local_service(self, channel):
            # TODO DiagnosticsService
            # return [DiagnosticsService(channel)]
            return []

        def get_service_proxy(self, channel, service_name):
            package_base = str(__package__) + ".remote"
            service = None
            try:
                # Xicom_v1.XX special case
                if 'Xicom_v1' in service_name:
                    service_name = "Xicom"
                clsName = service_name + "Proxy"
                clsName = clsName.replace('.', '_')
                package = package_base + "." + clsName
                clsModule = __import__(package, fromlist=[clsName],
                                       globals=globals())
                cls = clsModule.__dict__.get(clsName)
                service = cls(channel)
                # assert service_name == service.getName()
            except ImportError:
                pass
            except Exception as x:
                protocol.log("Cannot instantiate service proxy for " +
                             service_name, x)
            return service


    add_service_provider(DefaultServiceProvider())
